# Data

from __future__ import annotations

from pydantic import BaseModel, Field

from systembridgeshared.models.battery import Battery
from systembridgeshared.models.bridge import Bridge
from systembridgeshared.models.cpu import Cpu
from systembridgeshared.models.disk import Disk
from systembridgeshared.models.display import Display
from systembridgeshared.models.gpu import Gpu
from systembridgeshared.models.memory import Memory
from systembridgeshared.models.network import Network
from systembridgeshared.models.sensors import Sensors
from systembridgeshared.models.system import System


class Data(BaseModel):
    """Data"""

    battery: Battery = Field(..., alias="battery")
    bridge: Bridge = Field(..., alias="bridge")
    cpu: Cpu = Field(..., alias="cpu")
    disk: Disk = Field(..., alias="disk")
    display: Display = Field(..., alias="display")
    gpu: Gpu = Field(..., alias="gpu")
    memory: Memory = Field(..., alias="memory")
    network: Network = Field(..., alias="network")
    sensors: Sensors = Field(..., alias="sensors")
    system: System = Field(..., alias="system")



"""System Bridge: Settings"""
from __future__ import annotations

import io
import os
from os.path import exists
from uuid import uuid4

from appdirs import AppDirs
from cryptography.fernet import Fernet

from systembridgeshared.base import Base
from systembridgeshared.common import convert_string_to_correct_type
from systembridgeshared.const import (
    COLUMN_KEY,
    COLUMN_TIMESTAMP,
    COLUMN_VALUE,
    SECRET_API_KEY,
    SETTING_ADDITIONAL_MEDIA_DIRECTORIES,
    SETTING_AUTOSTART,
    SETTING_LOG_LEVEL,
    SETTING_PORT_API,
    TABLE_SECRETS,
    TABLE_SETTINGS,
)
from systembridgeshared.database import Database


class Settings(Base):
    """Settings"""

    def __init__(
        self,
        database: Database,
    ) -> None:
        """Initialize"""
        super().__init__()

        self._database = database
        self._database.create_table(
            TABLE_SECRETS,
            [
                (COLUMN_KEY, "TEXT PRIMARY KEY"),
                (COLUMN_VALUE, "TEXT"),
                (COLUMN_TIMESTAMP, "DOUBLE"),
            ],
        )

        self._database.create_table(
            TABLE_SETTINGS,
            [
                (COLUMN_KEY, "TEXT PRIMARY KEY"),
                (COLUMN_VALUE, "TEXT"),
                (COLUMN_TIMESTAMP, "DOUBLE"),
            ],
        )

        # Generate default encryption key
        self._encryption_key = None
        secret_key_path = os.path.join(
            AppDirs("systembridge", "timmo001").user_data_dir, "secret.key"
        )
        if exists(secret_key_path):
            with io.open(secret_key_path, encoding="utf-8") as file:
                self._encryption_key = file.read().splitlines()[0]
        if not self._encryption_key:
            self._encryption_key = Fernet.generate_key()
            with io.open(secret_key_path, "w", encoding="utf-8") as file:
                file.write(self._encryption_key.decode())

        # Default Secrets
        if self._database.check_table_for_key(TABLE_SECRETS, SECRET_API_KEY):
            self.set_secret(SECRET_API_KEY, str(uuid4()))

        # Default Settings
        if self._database.check_table_for_key(TABLE_SETTINGS, SETTING_AUTOSTART):
            self.set(SETTING_AUTOSTART, False)
        if self._database.check_table_for_key(TABLE_SETTINGS, SETTING_LOG_LEVEL):
            self.set(SETTING_LOG_LEVEL, "INFO")
        if self._database.check_table_for_key(TABLE_SETTINGS, SETTING_PORT_API):
            self.set(SETTING_PORT_API, 9170)
        if self._database.check_table_for_key(
            TABLE_SETTINGS, SETTING_ADDITIONAL_MEDIA_DIRECTORIES
        ):
            self.set(SETTING_ADDITIONAL_MEDIA_DIRECTORIES, [])

    def get_all(self) -> list[dict]:
        """Get settings"""
        records = self._database.read_table(TABLE_SETTINGS).to_dict(orient="records")
        for record in records:
            record[COLUMN_VALUE] = convert_string_to_correct_type(record[COLUMN_VALUE])
        return records

    def get(
        self,
        key: str,
    ) -> bool | float | int | str | list | dict | None:
        """Get setting"""
        record = self._database.read_table_by_key(TABLE_SETTINGS, key).to_dict(
            orient="records"
        )
        return (
            convert_string_to_correct_type(record[0]["value"])
            if record and len(record) > 0
            else None
        )

    def get_secret(
        self,
        key: str,
    ) -> str:
        """Get secret"""
        record = self._database.read_table_by_key(TABLE_SECRETS, key).to_dict(
            orient="records"
        )
        if not record or len(record) < 1:
            raise KeyError(f"Secret {key} not found")

        secret = record[0]["value"]
        fernet = Fernet(self._encryption_key)  # type: ignore
        return fernet.decrypt(secret.encode()).decode()

    def set(
        self,
        key: str,
        value: bool | float | int | str | list | dict | None,
    ) -> None:
        """Set setting"""
        self._database.write(TABLE_SETTINGS, key, value)

    def set_secret(
        self,
        key: str,
        value: str,
    ) -> None:
        """Set secret"""
        fernet = Fernet(self._encryption_key)  # type: ignore

        self._database.write(
            TABLE_SECRETS, key, fernet.encrypt(value.encode()).decode()
        )



# generated by datamodel-codegen:
#   filename:  disk.json

from __future__ import annotations

from pydantic import BaseModel, Extra, Field


class LastUpdated(BaseModel):
    """
    Last updated
    """

    class Config:
        extra = Extra.allow

    devices: float
    io_counters_read_count: float
    io_counters_write_count: float
    io_counters_read_bytes: float
    io_counters_write_bytes: float
    io_counters_read_time: float
    io_counters_write_time: float
    partitions: float


class Disk(BaseModel):
    """
    Disk
    """

    class Config:
        extra = Extra.allow

    devices: list
    io_counters_read_count: int
    io_counters_write_count: int
    io_counters_read_bytes: int
    io_counters_write_bytes: int
    io_counters_read_time: int
    io_counters_write_time: int
    partitions: list
    last_updated: LastUpdated = Field(..., description="Last updated")



"""System Bridge: Update"""
from __future__ import annotations

import asyncio
import importlib.util
from json import loads
import os
import platform
import subprocess
import sys
import urllib.request

from systembridgeshared.base import Base


class Update(Base):
    """Update"""

    def _install_package(
        self,
        package: str,
    ) -> None:
        """Install a package."""
        args = [
            sys.executable,
            "-m",
            "pip",
            "install",
            package,
        ]
        self._logger.debug(" ".join(args))
        with subprocess.Popen(args, stdout=subprocess.PIPE) as pipe:
            result = pipe.communicate()[0].decode()
        self._logger.info("Result: %s", result)

    def _update(self, packages: dict[str, str | None]) -> None:
        """Update each package."""
        for package, version in packages.items():
            self._logger.info("Updating %s to version %s", package, version)
            if version:
                self._install_package(f"{package}=={version}")

        asyncio.get_running_loop().call_later(
            5, os.execl, sys.executable, sys.executable, *sys.argv
        )

    def update(
        self,
        version: str | None,
        wait: bool = False,
    ) -> dict[str, str | None]:
        """Update the application."""
        packages = {
            "systembridgeshared": version,
            "systembridgebackend": version,
        }
        if importlib.util.find_spec("systembridgecli") is not None:
            packages["systembridgecli"] = version
        if importlib.util.find_spec("systembridgeconnector") is not None:
            packages["systembridgeconnector"] = version
        if importlib.util.find_spec("systembridgefrontend") is not None:
            packages["systembridgefrontend"] = version
        if importlib.util.find_spec("systembridgegui") is not None:
            packages["systembridgegui"] = version
        if (
            "Windows" in platform.system()
            and importlib.util.find_spec("systembridgewindowssensors") is not None
        ):
            packages["systembridgewindowssensors"] = version

        for package, new_version in packages.items():
            if new_version is None:
                with urllib.request.urlopen(
                    f"https://pypi.org/pypi/{package}/json"
                ) as response:
                    data = loads(response.read())
                new_version = data["info"]["version"]
                packages[package] = new_version

        if wait:
            self._update(packages)
        else:
            asyncio.get_running_loop().call_later(2, self._update, packages)

        return packages



"""System Bridge Shared: Database"""
from __future__ import annotations

from collections import OrderedDict
from collections.abc import Mapping
import json
import os
from sqlite3 import OperationalError, connect
from time import time

from pandas import DataFrame, read_sql_query

from systembridgeshared.base import Base
from systembridgeshared.common import (
    convert_string_to_correct_type,
    get_user_data_directory,
)
from systembridgeshared.const import (
    COLUMN_HARDWARE_NAME,
    COLUMN_HARDWARE_TYPE,
    COLUMN_KEY,
    COLUMN_NAME,
    COLUMN_TIMESTAMP,
    COLUMN_TYPE,
    COLUMN_VALUE,
)


class Database(Base):
    """Database"""

    def __init__(self):
        """Initialise"""
        super().__init__()
        self._connection = connect(
            os.path.join(get_user_data_directory(), "systembridge.db"),
            check_same_thread=False,
        )

    @property
    def connected(self) -> bool:
        """Check if connected"""
        return self._connection is not None

    def execute_sql(
        self,
        sql: str,
    ) -> None:
        """Execute SQL"""
        if not self.connected:
            self.connect()
        self._logger.debug("Executing SQL: %s", sql)
        try:
            self._connection.commit()
        except OperationalError:
            pass
        try:
            self._connection.execute(sql)
            self._connection.commit()
        except OperationalError as error:
            self._logger.error(
                "Error executing SQL: %s\n%s",
                sql,
                error,
            )

    def execute_sql_with_params(
        self,
        sql: str,
        params: Mapping,
    ) -> None:
        """Execute SQL"""
        self._logger.debug("Executing SQL: %s\n%s", sql, params)
        try:
            self._connection.commit()
        except OperationalError:
            pass
        try:
            self._connection.execute(sql, params)
            self._connection.commit()
        except OperationalError as error:
            self._logger.error(
                "Error executing SQL: %s\n%s",
                sql,
                error,
            )

    def connect(self) -> None:
        """Connect to database"""
        self._connection = connect(
            os.path.join(get_user_data_directory(), "systembridge.db"),
            check_same_thread=False,
        )

    def close(self) -> None:
        """Close connection"""
        self._connection.close()

    def check_table_for_key(
        self,
        table_name: str,
        key: str,
    ) -> bool:
        """Check if key exists in table"""
        return self.read_table_by_key(table_name, key).empty

    def read_table(
        self,
        table_name: str,
    ) -> DataFrame:
        """Read table"""
        return self.read_query(
            f"SELECT * FROM {table_name}",
        )

    def read_table_by_key(
        self,
        table_name: str,
        key: str,
    ) -> DataFrame:
        """Read table by key"""
        return self.read_query(
            f"SELECT * FROM {table_name} WHERE {COLUMN_KEY} = '{key}'",
        )

    def read_query(
        self,
        query: str,
    ) -> DataFrame:
        """Read SQL"""
        if not self.connected:
            self.connect()
        self._logger.debug("Reading SQL: %s", query)
        return read_sql_query(query, self._connection)

    def create_table(
        self,
        table_name: str,
        columns: list[tuple],
    ) -> None:
        """Create table"""
        sql = f"CREATE TABLE IF NOT EXISTS {table_name} ("
        for column in columns:
            sql += f"{column[0]} {column[1]},"
        sql = sql[:-1] + ")"
        self.execute_sql(sql)

    def clear_table(
        self,
        table_name: str,
    ) -> None:
        """Clear table"""
        self.execute_sql(f"DELETE FROM {table_name}")

    def clear_table_by_key(
        self,
        table_name: str,
        key: str,
    ) -> None:
        """Clear table by key"""
        self.execute_sql(f"DELETE FROM {table_name} WHERE {COLUMN_KEY} = '{key}'")

    def write(
        self,
        table_name: str,
        data_key: str,
        data_value: bool | float | int | str | list | dict | None,
        data_timestamp: float | None = None,
    ) -> None:
        """Write to table"""
        if data_timestamp is None:
            data_timestamp = time()

        # Convert list or dict to JSON
        if isinstance(data_value, (dict, list)):
            data_value = json.dumps(data_value)
        else:
            data_value = str(data_value)

        self.execute_sql(
            f"""INSERT INTO {table_name} ({COLUMN_KEY}, {COLUMN_VALUE}, {COLUMN_TIMESTAMP})
             VALUES ('{data_key}', '{data_value}', {data_timestamp})
             ON CONFLICT({COLUMN_KEY}) DO
             UPDATE SET {COLUMN_VALUE} = '{data_value}', {COLUMN_TIMESTAMP} = {data_timestamp}
             WHERE {COLUMN_KEY} = '{data_key}'
            """.replace(
                "\n", ""
            ).replace(
                "            ", ""
            ),
        )

    def write_sensor(
        self,
        table_name: str,
        data_key: str,
        data_type: str,
        data_name: str,
        data_hardware_type: str,
        data_hardware_name: str,
        data_value: bool | float | int | str | list | dict | None,
        data_timestamp: float | None = None,
    ) -> None:
        """Write to table"""
        if data_timestamp is None:
            data_timestamp = time()

        # Convert list or dict to JSON
        if isinstance(data_value, (dict, list)):
            data_value = json.dumps(data_value)
        else:
            data_value = str(data_value)

        self.execute_sql(
            f"""INSERT INTO {table_name} ({COLUMN_KEY}, {COLUMN_TYPE}, {COLUMN_NAME},
             {COLUMN_HARDWARE_TYPE}, {COLUMN_HARDWARE_NAME}, {COLUMN_VALUE}, {COLUMN_TIMESTAMP})
             VALUES ('{data_key}', '{data_type}', '{data_name}', '{data_hardware_type}',
             '{data_hardware_name}', '{data_value}', {data_timestamp})
             ON CONFLICT({COLUMN_KEY}) DO
             UPDATE SET {COLUMN_VALUE} = '{data_value}', {COLUMN_TIMESTAMP} = {data_timestamp}
             WHERE {COLUMN_KEY} = '{data_key}'
            """.replace(
                "\n", ""
            ).replace(
                "            ", ""
            ),
        )

    def table_data_to_ordered_dict(
        self,
        table_name: str,
    ) -> OrderedDict | None:
        """Convert table to OrderedDict"""
        data_dict = self.read_table(table_name).to_dict(orient="records")
        if len(data_dict) == 0:
            return None
        data: dict = {
            "last_updated": {},
        }
        for item in data_dict:
            data = {
                **data,
                item[COLUMN_KEY]: convert_string_to_correct_type(item[COLUMN_VALUE]),
                "last_updated": {
                    **data["last_updated"],  # type: ignore
                    item[COLUMN_KEY]: item["timestamp"],
                },
            }
        output = OrderedDict(sorted(data.items()))
        output.move_to_end("last_updated", last=True)

        return output



"""System Bridge Shared: Constants"""
from systembridgeshared.models.battery import Battery
from systembridgeshared.models.bridge import Bridge
from systembridgeshared.models.cpu import Cpu
from systembridgeshared.models.data import Data
from systembridgeshared.models.disk import Disk
from systembridgeshared.models.display import Display
from systembridgeshared.models.generic import Generic
from systembridgeshared.models.gpu import Gpu
from systembridgeshared.models.keyboard_key import KeyboardKey
from systembridgeshared.models.keyboard_text import KeyboardText
from systembridgeshared.models.media_directories import MediaDirectories
from systembridgeshared.models.media_files import File, MediaFiles
from systembridgeshared.models.memory import Memory
from systembridgeshared.models.network import Network
from systembridgeshared.models.notification import Notification
from systembridgeshared.models.open_path import OpenPath
from systembridgeshared.models.open_url import OpenUrl
from systembridgeshared.models.response import Response
from systembridgeshared.models.sensors import Sensors
from systembridgeshared.models.system import System

# Logging
DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
FORMAT = "%(asctime)s %(levelname)s (%(threadName)s) [%(name)s] %(message)s"

# Database
TABLE_SECRETS = "secrets"
TABLE_SETTINGS = "settings"

COLUMN_HARDWARE_NAME = "hardware_name"
COLUMN_HARDWARE_TYPE = "hardware_type"
COLUMN_KEY = "key"
COLUMN_NAME = "name"
COLUMN_TIMESTAMP = "timestamp"
COLUMN_TYPE = "type"
COLUMN_VALUE = "value"

# Settings
SETTING_ADDITIONAL_MEDIA_DIRECTORIES = "additional_media_directories"
SETTING_AUTOSTART = "autostart"
SETTING_LOG_LEVEL = "log_level"
SETTING_PORT_API = "port_api"

# Secrets
SECRET_API_KEY = "api_key"

# Query Parameters
QUERY_ALBUM = "album"
QUERY_API_KEY = "apiKey"
QUERY_API_PORT = "apiPort"
QUERY_ARTIST = "artist"
QUERY_AUTOPLAY = "autoplay"
QUERY_BASE = "base"
QUERY_FILENAME = "filename"
QUERY_PATH = "path"
QUERY_TITLE = "title"
QUERY_URL = "url"
QUERY_VOLUME = "volume"

# Event Keys
EVENT_API_KEY = "api-key"
EVENT_APP_ICON = "app_icon"
EVENT_APP_NAME = "app_name"
EVENT_BASE = "base"
EVENT_DATA = "data"
EVENT_DIRECTORIES = "directories"
EVENT_EVENT = "event"
EVENT_FILE = "file"
EVENT_FILENAME = "filename"
EVENT_FILES = "files"
EVENT_ID = "id"
EVENT_KEY = "key"
EVENT_MESSAGE = "message"
EVENT_MODULE = "module"
EVENT_MODULES = "modules"
EVENT_PATH = "path"
EVENT_SETTING = "setting"
EVENT_SUBTYPE = "subtype"
EVENT_TEXT = "text"
EVENT_TIMEOUT = "timeout"
EVENT_TITLE = "title"
EVENT_TYPE = "type"
EVENT_URL = "url"
EVENT_VALUE = "value"
EVENT_VERSION = "version"
EVENT_VERSIONS = "versions"

# Event Types
SUBTYPE_BAD_API_KEY = "BAD_API_KEY"
SUBTYPE_BAD_DIRECTORY = "BAD_DIRECTORY"
SUBTYPE_BAD_FILE = "BAD_FILE"
SUBTYPE_BAD_JSON = "BAD_JSON"
SUBTYPE_BAD_KEY = "MISSING_KEY"
SUBTYPE_BAD_PATH = "BAD_PATH"
SUBTYPE_BAD_REQUEST = "BAD_REQUEST"
SUBTYPE_LISTENER_ALREADY_REGISTERED = "LISTENER_ALREADY_REGISTERED"
SUBTYPE_LISTENER_NOT_REGISTERED = "LISTENER_NOT_REGISTERED"
SUBTYPE_MISSING_API_KEY = "MISSING_API_KEY"
SUBTYPE_MISSING_BASE = "MISSING_BASE"
SUBTYPE_MISSING_KEY = "MISSING_KEY"
SUBTYPE_MISSING_MESSAGE = "MISSING_MESSAGE"
SUBTYPE_MISSING_MODULES = "MISSING_MODULES"
SUBTYPE_MISSING_PATH = "MISSING_PATH"
SUBTYPE_MISSING_PATH_URL = "MISSING_PATH_URL"
SUBTYPE_MISSING_SETTING = "MISSING_SETTING"
SUBTYPE_MISSING_TEXT = "MISSING_TEXT"
SUBTYPE_MISSING_VALUE = "MISSING_VALUE"
SUBTYPE_UNKNOWN_EVENT = "UNKNOWN_EVENT"
TYPE_APPLICATION_UPDATE = "APPLICATION_UPDATE"
TYPE_APPLICATION_UPDATING = "APPLICATION_UPDATING"
TYPE_DATA_GET = "DATA_GET"
TYPE_DATA_LISTENER_REGISTERED = "DATA_LISTENER_REGISTERED"
TYPE_DATA_LISTENER_UNREGISTERED = "DATA_LISTENER_UNREGISTERED"
TYPE_DATA_UPDATE = "DATA_UPDATE"
TYPE_DIRECTORIES = "DIRECTORIES"
TYPE_ERROR = "ERROR"
TYPE_EXIT_APPLICATION = "EXIT_APPLICATION"
TYPE_FILE = "FILE"
TYPE_FILES = "FILES"
TYPE_GET_DATA = "GET_DATA"
TYPE_GET_DIRECTORIES = "GET_DIRECTORIES"
TYPE_GET_FILE = "GET_FILE"
TYPE_GET_FILES = "GET_FILES"
TYPE_GET_SETTING = "GET_SETTING"
TYPE_GET_SETTINGS = "GET_SETTINGS"
TYPE_KEYBOARD_KEY_PRESSED = "KEYBOARD_KEY_PRESSED"
TYPE_KEYBOARD_KEYPRESS = "KEYBOARD_KEYPRESS"
TYPE_KEYBOARD_TEXT = "KEYBOARD_TEXT"
TYPE_KEYBOARD_TEXT_SENT = "KEYBOARD_TEXT_SENT"
TYPE_NOTIFICATION = "NOTIFICATION"
TYPE_NOTIFICATION_SENT = "NOTIFICATION_SENT"
TYPE_OPEN = "OPEN"
TYPE_OPENED = "OPENED"
TYPE_POWER_HIBERNATE = "POWER_HIBERNATE"
TYPE_POWER_HIBERNATING = "POWER_HIBERNATING"
TYPE_POWER_LOCK = "POWER_LOCK"
TYPE_POWER_LOCKING = "POWER_LOCKING"
TYPE_POWER_LOGGINGOUT = "POWER_LOGGINGOUT"
TYPE_POWER_LOGOUT = "POWER_LOGOUT"
TYPE_POWER_RESTART = "POWER_RESTART"
TYPE_POWER_RESTARTING = "POWER_RESTARTING"
TYPE_POWER_SHUTDOWN = "POWER_SHUTDOWN"
TYPE_POWER_SHUTTINGDOWN = "POWER_SHUTTINGDOWN"
TYPE_POWER_SLEEP = "POWER_SLEEP"
TYPE_POWER_SLEEPING = "POWER_SLEEPING"
TYPE_REGISTER_DATA_LISTENER = "REGISTER_DATA_LISTENER"
TYPE_SETTING_RESULT = "SETTING_RESULT"
TYPE_SETTING_UPDATED = "SETTING_UPDATED"
TYPE_SETTINGS_RESULT = "SETTINGS_RESULT"
TYPE_UNREGISTER_DATA_LISTENER = "UNREGISTER_DATA_LISTENER"
TYPE_UPDATE_SETTING = "UPDATE_SETTING"

# Model
MODEL_BATTERY = "battery"
MODEL_BRIDGE = "bridge"
MODEL_CPU = "cpu"
MODEL_DATA = "data"
MODEL_DISK = "disk"
MODEL_DISPLAY = "display"
MODEL_GENERIC = "generic"
MODEL_GPU = "gpu"
MODEL_KEYBOARD_KEY = "keyboard_key"
MODEL_KEYBOARD_TEXT = "keyboard_text"
MODEL_MEDIA_DIRECTORIES = "media_directories"
MODEL_MEDIA_FILES = "media_files"
MODEL_MEDIA_FILE = "media_file"
MODEL_MEMORY = "memory"
MODEL_NETWORK = "network"
MODEL_NOTIFICATION = "notification"
MODEL_OPEN_PATH = "open_path"
MODEL_OPEN_URL = "open_url"
MODEL_RESPONSE = "response"
MODEL_SENSORS = "sensors"
MODEL_SYSTEM = "system"

MODEL_MAP = {
    MODEL_BATTERY: Battery,
    MODEL_BRIDGE: Bridge,
    MODEL_CPU: Cpu,
    MODEL_DATA: Data,
    MODEL_DISK: Disk,
    MODEL_DISPLAY: Display,
    MODEL_GENERIC: Generic,
    MODEL_GPU: Gpu,
    MODEL_KEYBOARD_KEY: KeyboardKey,
    MODEL_KEYBOARD_TEXT: KeyboardText,
    MODEL_MEDIA_DIRECTORIES: MediaDirectories,
    MODEL_MEDIA_FILE: File,
    MODEL_MEDIA_FILES: MediaFiles,
    MODEL_MEMORY: Memory,
    MODEL_NETWORK: Network,
    MODEL_NOTIFICATION: Notification,
    MODEL_OPEN_PATH: OpenPath,
    MODEL_OPEN_URL: OpenUrl,
    MODEL_RESPONSE: Response,
    MODEL_SENSORS: Sensors,
    MODEL_SYSTEM: System,
}
